#ifndef LINUX_NBD_H
#define LINUX_NBD_H

/* unsigned comments are Pavel's originals for 2.1.*
 *   pavel@atrey.karlin.mff.cuni.cz (Pavel Machek)
 * comments marked PTB are from
 *   ptb@it.uc3m.es (Peter T. Breuer)
 * comments marked AMARIN are from
 *   amarin@it.uc3m.es (Andres Marin Lopez)
 */

#include <asm/types.h>

/* PTB - old style ioctl assignments
#define NBD_SET_SOCK    0x4E00
#define NBD_SET_BLKSIZE 0x4E01
#define NBD_SET_SIZE    0x4E02
#define NBD_DO_IT       0x4E03
#define NBD_CLEAR_SOCK  0x4E04
#define NBD_CLEAR_QUE   0x4E05
#define NBD_PRINT_DEBUG 0x4E06
#define NBD_GET_BLKSIZE 0x4E07
#define NBD_HARD_RESET  0x4E08
#define NBD_DEC_USE_COUNT 0x4E09
#define MY_NBD_ACK      0x4E0a
#define MY_NBD_GET_REQ  0x4E0b
#define MY_NBD_REG_BUF  0x4E0c
#define MY_NBD_CLR_REQ  0x4E0d
#define MY_NBD_SYNC     0x4E0e
#define NBD_SET_SECTORS 0x4E0f
#define MY_NBD_SET_SIG  0x4E10
#define NBD_RESET       0x4E11
#define NBD_SET_RO      0x4E12
#define MY_NBD_ERR_REQ  0x4E13
 */

/* PTB - new style ioctl assignments */
#define NBD_SET_SOCK    _IOW(0xab, 0, int)
#define NBD_SET_BLKSIZE _IOW(0xab, 1, int)
#define NBD_SET_SIZE    _IOW(0xab, 2, int)
#define NBD_DO_IT       _IOW(0xab, 3, int)
#define NBD_CLEAR_SOCK  _IOW(0xab, 4, int)
#define NBD_CLEAR_QUE   _IO(0xab, 5)
#define NBD_PRINT_DEBUG _IO(0xab, 6)
#define NBD_GET_BLKSIZE _IOR(0xab, 7, long *)
#define NBD_HARD_RESET  _IO(0xab, 8)
#define NBD_DEC_USE_COUNT _IO(0xab, 9)
#define MY_NBD_ACK      _IOW(0xab, 10, char *)
#define MY_NBD_GET_REQ  _IOR(0xab, 11, char *)
#define MY_NBD_REG_BUF  _IOW(0xab, 12, char *)
#define MY_NBD_CLR_REQ  _IOW(0xab, 13, int)
#define MY_NBD_SYNC     _IOW(0xab, 14, int)
#define NBD_SET_SECTORS _IOW(0xab, 15, int)
#define MY_NBD_SET_SIG  _IOW(0xab, 16, int *)
#define NBD_RESET       _IO(0xab, 17)
#define NBD_SET_RO      _IOW(0xab, 18, int)
#define MY_NBD_ERR_REQ  _IOW(0xab, 19, int)
 
#define MAX_NBD 16         /* PTB MAX was 128, but that's a lot */
#define NBD_SHIFT 4        /* PTB 16 partitions/sockets/slots per device */
                           /* PTB number of socket slots per device */
#define NBD_MAXCONN (1<<NBD_SHIFT)
                           /* PTB length of sig on device */
#define NBD_SIGLEN 128
#define NBD_MAX_SECTORS 168/* PTB max number of 512B sectors in a buffer */

#ifdef MAJOR_NR
  /* we are included from the kernel nbd.c file so put kernel stuff here PTB */

  #include <linux/locks.h>
  #include <linux/config.h>

  #define ENDREQ_NOCURRENT
  #define LOCAL_END_REQUEST
  #include <linux/blk.h>

  #if LINUX_VERSION_CODE > LINUX_VERSION_2_1_0
  static void end_request(struct request *req, int uptodate) {  

        unsigned long flags;

        spin_lock_irqsave(&io_request_lock, flags);
        while (end_that_request_first(req, uptodate, DEVICE_NAME));
    #ifndef DEVICE_NO_RANDOM
        add_blkdev_randomness(MAJOR(req->rq_dev)); 
    #endif
        /* DEVICE_OFF(req->rq_dev); // PTB 2.2.21 */
        /* CURRENT = req->next; // PTB 2.2.21 */
        end_that_request_last(req);     
        spin_unlock_irqrestore(&io_request_lock, flags);
  }                 
  #endif


  static void nbd_end_request(struct request *req) {
   /* this is elaborated in 2.1.132 but I don't think there's support
    * or need for that elaboration in 2.0.* PTB
    */
   req->cmd &= 0x03; /* PTB */
   /* unlock chained buffers */
   while (req && req->nr_sectors > req->current_nr_sectors){
            req->nr_sectors -= req->current_nr_sectors;
            req->sector     += req->current_nr_sectors;
            end_request (req, !req->errors);
   }
   end_request( req, !req->errors );
  }



  /* PTB various defaults */
  #define NBD_QMAX    2      /* PTB queue limit enforced by drive semaphore */
  #define NBD_CHEAT_MIN  30  /* PTB read twice within 30-70 jiffies to reset */
  #define NBD_CHEAT_MAX  70  /* PTB read twice within 30-70 jiffies to reset */
  #define NBD_RAHEAD 24      /* PTB slow medium                  */
  #define NBD_SYNC_INTVL 30  /* PTB client inactivity check interval */
  #define NBD_MERGE_REQUESTS 0 /* PTB until accounting fixed! */
  #define NBD_AGE_REQ 5      /* PTB secs after which a request is rolled back */

  struct nbd_slot {
    struct semaphore lock;   /* PTB socket                       */
    struct file * file;      /* PTB add - for refcnt, NULL if slot empty */
    struct socket * sock;    /* PTB add                          */
    int in;                  /* PTB add - tot blocks pended      */
    int out;                 /* PTB add - tot blocks released    */
    int err;                 /* PTB add - tot blocks errored     */
    char * buffer;           /* PTB add - user space buffer      */
    int  bufsiz;             /* PTB add - user space buffer size */
    #if LINUX_VERSION_CODE < LINUX_VERSION_2_3_30
    struct request * tail;    /* PTB add - rollback request       */
    #else
    struct list_head queue;
    #endif
    unsigned long req_age;   /* PTB add - age of pending req     */
    unsigned long cli_age;   /* PTB add - age of client          */
    struct nbd_device *lo;   /* PTB add - parent device          */
  #define NBD_SLOT_RUNNING   0x01
  #define NBD_SLOT_WAITING   0x02
  #define NBD_SLOT_BUFFERED  0x04
    int flags;               /* PTB add */
    int i;                   /* PTB add - slot number */
    int buflen;              /* PTB add - buffer byte count */
    int pid;                 /* PTB add - client process */
    #if LINUX_VERSION_CODE < LINUX_VERSION_2_3_30
    struct request * head;    /* PTB add - rollback request       */
    #endif
  };

  struct nbd_device {
    int refcnt;	
  #define NBD_READ_ONLY   0x0001
  #define NBD_WRITE_NOCHK 0x0002
  #define NBD_INITIALISED 0x0004
  #define NBD_SIGNED      0x0008
  #define NBD_ENABLED     0x0010
  #define NBD_SIZED       0x0020
  #define NBD_BLKSIZED    0x0040
    int flags;
    int harderror;		/* Code of hard error			*/
    int magic;			/* FIXME: not if debugging is off	*/
  #if LINUX_VERSION_CODE > LINUX_VERSION_2_3_30
    struct list_head queue;
  #else
    struct request *head;	        /* Requests are added here...		*/
    struct request *tail;         /* PTB - and usually removed here */
  #endif
    struct semaphore queue_lock;          /* PTB add - queue access */
    struct semaphore depth_lock;          /* PTB add - max queue depth   */
    int nslot;                            /* PTB add - total slots */
    int islot;                            /* PTB add - current slot */
    int aslot;                            /* PTB add - total active slots*/
    int requests_in;                      /* PTB add - blocks placed on queue */
    int requests_out;                     /* PTB add - blocks taken off queue */
    int requests_err;                     /* PTB add - blocks errored off queue */
    int requests_req;                     /* PTB add - blocks pending */
    int kwaiters;                         /* PTB add - kernel threads waiting */
    int kthreads;                         /* PTB add - kernel threads in */
    int maxq;                             /* PTB add - max request queue depth */
    int count;                            /* PTB add - request queue depth */
    int errors;                           /* PTB add - tot requests errored */
    int seqno;                            /* PTB add - sequence number */
    int cwaiters;                         /* PTB add - client threads waiting */
    int cthreads;                         /* PTB add - client threads in */
  #if LINUX_VERSION_CODE > LINUX_VERSION_2_3_30
    wait_queue_head_t wq;                 /* PTB add */
  #else
    struct wait_queue* wq;                /* AMARIN/PTB add */
  #endif
    struct nbd_slot slots[NBD_MAXCONN];   /* PTB add - client array */
    unsigned blksize;                     /* PTB add - device blksize in B */
    u64 bytesize;                         /* PTB add - device size in B */
    unsigned sectors;                     /* PTB add - nr 512B sectors */
    unsigned size;                        /* PTB add - device size in KB */
    unsigned logblksize;                  /* PTB add - log2 blksize */
    unsigned nbd;                         /* PTB add - this array index */
    int signature[NBD_SIGLEN/sizeof(int)];
                                          /* PTB add - server sig */
    struct file * file;                   /* PTB add - for ref */
    struct inode * inode;                 /* PTB add - for ref */
    int  bufsiz;                          /* PTB add - user space buffer size */
    int kmax;                             /* PTB add - max kernel threads */
    char *blockmap;                       /* PTB add - map of block states */
    unsigned long disabled;               /* PTB add - when was it disabled */
  };

#endif  /* MAJOR_NR */



/* This now IS in some kind of include file...	*/

/* PTB 132 */ 
#define NBD_INIT_MAGIC 0x12345678       /* AMARIN */
#define NBD_REQUEST_MAGIC 0x25609513
#define NBD_REPLY_MAGIC 0x67446698     
/* Do *not* use magics: 0x12560953 0x96744668. 
 */

#define NBD_DEV_MAGIC 0x68797548

#define NBD_REQUEST_MAGIC_T  __u32
#define NBD_REQUEST_TYPE_T   __u32
#define NBD_HANDLE_LENGTH    sizeof(void*)      /* PTB number of bytes */
#define NBD_REQUEST_FROM_T   __u64
#define NBD_REQUEST_LEN_T    __u32
#define NBD_REQUEST_FLAGS_T  __u32
#define NBD_REPLY_MAGIC_T    __u32
#define NBD_REPLY_ERROR_T    __u32
#define NBD_REPLY_FLAGS_T    __u32

typedef char nbd_handle_t[NBD_HANDLE_LENGTH];

#define NBD_REQUEST_HANDLE_T nbd_handle_t
#define NBD_REPLY_HANDLE_T   nbd_handle_t

struct nbd_request {
  NBD_REQUEST_MAGIC_T  magic;
  NBD_REQUEST_TYPE_T   type;			/* == READ || == WRITE 	*/
  NBD_REQUEST_HANDLE_T handle;
  NBD_REQUEST_FROM_T   from;                    /* 64 bit PTB 132 */
  NBD_REQUEST_LEN_T    len;
#define NBD_REQUEST_CACHED     0x0100
#define NBD_REQUEST_BITMAPPED  0x0200
#define NBD_REQUEST_REMOTE     0x0400
#define NBD_REQUEST_ERRORED    0x0800
  NBD_REQUEST_FLAGS_T  flags;
} /* __attribute__ ((packed)) */;

struct nbd_reply {
  NBD_REPLY_MAGIC_T    magic;
  NBD_REPLY_ERROR_T    error;		      /* 0 = ok, else error	*/
  NBD_REPLY_HANDLE_T   handle;	              /* handle you got from request */
#define NBD_REPLY_CACHED       0x0100
#define NBD_REPLY_BITMAPPED    0x0200
#define NBD_REPLY_REMOTE       0x0400
#define NBD_REPLY_ERRORED      0x0800
  NBD_REPLY_FLAGS_T    flags;
} /* __attribute__ ((packed)) */;


#endif /* LINUX_NBD_H */


